/*============================================================================
  File:     PSP and UNKNOWN.sql

  Summary:  Understanding how PSP problems occur and how to avoid SOME
            using optimize for unknown (was new in SQL Server 2008).

  Date:     March 2012

  SQL Server Version: SQL Server 200x
------------------------------------------------------------------------------
  Written by Kimberly L. Tripp, SYSolutions, Inc.

  For more scripts and sample code, check out 
    http://www.SQLskills.com

  This script is intended only as a supplement to demos and lectures
  given by Kimberly L. Tripp.  
  
  THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF 
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED 
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
============================================================================*/

USE CREDIT
go

UPDATE member 
    SET lastname = 'tripp' 
    WHERE member_no = 123
go

CREATE INDEX MemberName ON Member(Lastname)
go

DBCC SHOW_STATISTICS (member, membername)
WITH HISTOGRAM
go

-- This proc is a typical proc...
CREATE PROC p1
    @lname  varchar(15)
AS
SELECT * 
FROM dbo.member AS m
WHERE m.lastname = @lname
go

-- On first execution, SQL Server uses parameter sniffing...
-- This is GOOD for the first execution as the 
-- estimate is correct and SQL Server chooses 
-- the right plan for this value.
exec p1 'tripp'
go

-- However, what if you supply a non-standard value?
-- Since the plan was already determined by the first execution
-- then we're going to get THAT plan. Is that good for this
-- value? Maybe... but, maybe not? This is where parameter
-- sniffing becomes a problem!
exec p1 'anderson'
go

-- One solution that people found is to "obfuscate" the
-- parameter coming in...

-- This procedure uses a variable inside the procedure 
-- to "trick" SQL Server at optimization/compilation time.
-- This turns the parameter into an "unknown" value and
-- as a result - SQL Server cannot "sniff" the incoming
-- value.
CREATE PROC p2
    @lname  varchar(15)
AS
DECLARE @lastname   varchar(15)
SELECT @lastname = @lname
SELECT * 
FROM dbo.member AS m
WHERE m.lastname = @lastname
go

exec p2 'tripp'
go
-- estimate is incorrect. SQL Server chose a
-- plan based on the density vector (the all density)
-- which is the "AVERAGE"

exec p2 'anderson'
go
-- Here we get a "better" plan for this parameter...

-- What SQL Server is doing is getting the "average" plan...
-- If we actually look at the data, you'll find that 'Tripp'
-- is the anomalie. So, it might actually be preferred to
-- use the "average" plan. To simplify this process, they
-- added a new statement hint in SQL Server 2008.

-- In SQL Server 2008
CREATE PROC p3
    @lname  varchar(15)
AS
SELECT * 
FROM dbo.member AS m
WHERE m.lastname = @lname
OPTION (OPTIMIZE FOR UNKNOWN)
go

exec p3 'tripp'
go

exec p3 'anderson'
go

-- My issue is that there are cases where the average plan
-- is good enough. But, you need to be careful. There are 
-- some other options...

-- OPTION (RECOMPILE)
-- OPTION (OPTIMIZE FOR ( ))

-- Finally, there is a way to disable parameter sniffing 
-- SERVER-WIDE (not recommended) using Trace Flag 4136

-- http://support.microsoft.com/kb/980653 

-- If this trace flag is on then parameter sniffing will be
-- turned off and all code will use the density vector
-- for optimization. It's as if you added OPTIMIZE FOR UNKNOWN
-- to ALL of your code.